package decrypto;

import java.util.*;

/** A plaintext word that might correspond to a particular
 * ciphertext. Each Word contains a list of candidates as found from a
 * dictionary. 
 **/
public class Candidate
{
    /** This score is used to rank candidates for the purposes of
	searching. It is not used for scoring solutions (in general,
	at least, because trigrams span multiple words.) 
    **/
    public double score;
    public byte[] cand;

    public Candidate(byte[] cand, double score)
    {
	this.cand = cand;
	this.score = score;
    }

    public static Candidate[] toCandidates(ArrayList<byte[]> cands, Language lang)
    {
	Candidate c[] = new Candidate[cands.size()];
	for (int i = 0; i < cands.size(); i++)
	    c[i] = new Candidate(cands.get(i), lang.computeCandidateScore(cands.get(i)));

	Arrays.sort(c, new CandidateScoreComparator());
	return c;
    }

    // High scores first, low scores last
    public static class CandidateScoreComparator implements Comparator<Candidate>
    {
	public int compare(Candidate a, Candidate b)
	{
	    return -Double.compare(a.score, b.score);
	}
    }
}
