package decrypto;

import java.util.*;
import java.io.*;

import decrypto.statistics.*;

public class EnglishLanguage extends Language
{
    static float _tetragrams[][][][];
    static float _tetragramsNS[][][][];

    static Object syncObject = new Object();
    static 
    {
	new PreloadThread().start();
    }
    
    public EnglishLanguage()
    {
    }

    static class PreloadThread extends Thread
    {
	PreloadThread()
	{
	    setPriority(Thread.MIN_PRIORITY);
	}

	public void run()
	{
	    getNgramsStatic(4, true);
	    getNgramsStatic(4, false);
	}
    }

    public Object getNgrams(int order, boolean withSpaces) 
    {
	return getNgramsStatic(order, withSpaces);
    }

    static Object getNgramsStatic(int order, boolean withSpaces) 
    {
	synchronized (syncObject) {
	    try {
		if (order==3) {
		    assert(false);
		}

		if (order==4) {
		    if (withSpaces) {
			if (_tetragrams == null)
			    _tetragrams = BuildStatistics4.readBinary("english4.sta");
			return _tetragrams;
		    } else {
			if (_tetragramsNS == null)
			    _tetragramsNS = BuildStatistics4.readBinary("english4ns.sta");
			return _tetragramsNS;
		    }
		}
	    } catch (IOException ex) {
		System.out.println("Couldn't open statistics: "+ex);
	    }
	}
	return null;
    }

    public char byteToChar(byte v)
    {
	if (v==-1)
	    return '~';

	if (v==26)
	    return ' ';

	return (char) ('A'+v);
    }

    public final byte charToByte(char c)
    {
	if (c==' ')
	    return (byte) letterCount();

	c = Character.toUpperCase(c);

	if (c<'A' || c>'Z')
	    return (byte) letterCount();

	return (byte) (c-'A');
    }

    public int letterCount()
    {
	return 26;
    }

    public String translateString(String ciphertext, MapSet map, int spaces[])
    {
	StringBuffer sb = new StringBuffer();
	int spacepos = 0;

	// we delay spaces until we hit the next alphabetic character:
	// this allows punctuation to immediately follow
	// letters. I.e., we get "Once upon a time, there..." instead
	// of "Once upon a time ,there..."
	boolean spacepending = false; 

	for (int i = 0; i < ciphertext.length(); i++) {
	    char c = ciphertext.charAt(i);

	    if (spaces != null && c==' ')
		continue;

	    byte cb = charToByte(c);

	    if (cb>=0 && cb < letterCount()) {
		if (spacepending) {
		    sb.append(' ');
		    spacepending = false;
		}

		int mc = map.getMapping(charToByte(c));
		if (mc == -1) {
		    sb.append("~");
		} else {
		    if (Character.isLowerCase(c))
			sb.append((char) ('a'+mc));
		    else 
			sb.append((char) ('A'+mc));
		}

		if (spaces != null && spaces[spacepos]>0)
		    spacepending = true;
		spacepos++;
	    } else {
		sb.append(c);
	    }
	}
	
	return sb.toString();
    }

    static final double unknownLetterPenalty = -20; //-20;

    public double computeCandidateScore(byte[] candidate)
    {
	int ma = 27; // unknown char
	int mb = 27; // unknown char
	int mc = 26; // a space
	double score = 0;

	float ngrams[][][][] = (float[][][][]) getNgrams(4, true);

 	for (int i = 0; i < candidate.length; i++)
	    {
		int md = (int) candidate[i];

		score += ngrams[ma][mb][mc][md];
		ma = mb;
		mb = mc;
		mc = md;
	    }

	score += ngrams[ma][mb][mc][26]; // add a space at the end.

	return score;
    }

    public double computeScore(byte[] cipherbytes, MapSet mapset)
    {
	int map[] = new int[28];
	for (int i = 0; i < 26; i++)
	    map[i] = mapset.getMapping((byte) i);
	map[26] = 26;
	map[27] = 27;

	int len = cipherbytes.length;
	int c0 = len > 2 ? cipherbytes[len-2] : 27;
	int c1 = len > 1 ? cipherbytes[len-1] : 27;
	int c2 = 26;

	int m0 = map[c0];
	if (m0 < 0)
	    m0 = 27;
	int m1 = map[c1];
	if (m1 < 0)
	    m1 = 27;
	int m2 = map[c2];
	if (m2 < 0)
	    m2 = 27;

	float grams[][][][] = (float[][][][]) getNgrams(4, true);

	double score = 0;

	for (int i = 0; i < cipherbytes.length; i++) {
	    int c3 = cipherbytes[i];
	    int m3 = map[c3];
	    if (m3 < 0) {
		m3 = 27;
		score += unknownLetterPenalty;
	    }

	    score += grams[m0][m1][m2][m3];
	    //	    System.out.printf("%3d %15f %c%c%c%c %15f\n", i, grams[m0][m1][m2][m3], m0+'A', m1+'A', m2+'A', m3+'A', score);

	    m0 = m1;
	    m1 = m2;
	    m2 = m3;
	}

	return score;
    }

    public double computeScore(byte[] cipherbytes, int map[])
    {
	int len = cipherbytes.length;
	int c0 = len > 2 ? cipherbytes[len-2] : 27;
	int c1 = len > 1 ? cipherbytes[len-1] : 27;
	int c2 = 26;

	int m0 = c0>=26 ? c0 : map[c0];
	if (m0 < 0)
	    m0 = 27;

	int m1 = c1>=26 ? c1 : map[c1];
	if (m1 < 0)
	    m1 = 27;

	int m2 = c2>=26 ? c2 : map[c2];
	if (m2 < 0)
	    m2 = 27;

	float grams[][][][] = (float[][][][]) getNgrams(4, true);

	double score = 0;

	for (int i = 0; i < cipherbytes.length; i++) {
	    int c3 = cipherbytes[i];
	    int m3 = c3==26 ? c3 : map[c3];
	    if (m3 < 0) {
		m3 = 27;
		score += unknownLetterPenalty;
	    }

	    score += grams[m0][m1][m2][m3];
	    //	    System.out.printf("%3d %15f %c%c%c%c\n", i, grams[m0][m1][m2][m3], m0+'A', m1+'A', m2+'A', m3+'A');

	    m0 = m1;
	    m1 = m2;
	    m2 = m3;
	}

	return score;
    }
}
