package decrypto;

import java.io.*;
import java.util.*;

import decrypto.dictattack.*;
import decrypto.genetic.*;

public class GABench
{
    public static void main(String args[])
    {
	try {
	    mainEx(args);
	} catch (IOException ex) {
	    System.out.println("ex: "+ex);
	}
    }

    static class Solution
    {
	double score = -Double.MAX_VALUE;
	String plaintext;
	double time;
	boolean solved;
    }

    static class MyDecryptoListener implements SolverListener
    {
	Solution bestSolution = new Solution();
	boolean finished = false;
	double scorethresh;
	Solver solver;
	Puzzle puzzle;
	TestCase tc;

	public void solverMessage(String s)
	{
	}

	public void solverFinished(double elapsedTime)
	{
	    synchronized(this) {
		finished = true;
		notifyAll();
	    }
	}

	public void solverSolution(MapSet mapset, int spaces[], boolean fullSolution, double score)
	{
	    int correctMappings = 0;
	    for (int i = 0; i < tc.map.length; i++) {
		if (mapset.getMapping((byte) i)==tc.map[i])
		    correctMappings++;
	    }

	    if (score < bestSolution.score)
		return;

	    bestSolution.score = score;
	    bestSolution.plaintext = puzzle.lang.translateString(puzzle.cipherText, mapset, spaces);

	    //	    System.out.println(correctMappings+" "+puzzle.numUniqueCiphertextLetters+" "+bestSolution.plaintext);
	    //if (bestSolution.score > scorethresh)
	    if ( bestSolution.score >= scorethresh) {
		//	    if (correctMappings >= puzzle.numUniqueCiphertextLetters - 3 || bestSolution.score > scorethresh) {
		bestSolution.time = solver.getElapsedTime();
		bestSolution.solved = true;
		solver.stop();
	    }
	}

	public void solverProgress(double progress)
	{
	}
    }

    static Solution solveWait(Puzzle puzzle, Solver solver, TestCase tc, double timelimit, double scoreThresh)
    {
	MyDecryptoListener mdl = new MyDecryptoListener();
	mdl.puzzle = puzzle;
	mdl.solver = solver;
	mdl.scorethresh = scoreThresh;
	mdl.tc = tc;

	solver.addListener(mdl);
	solver.start(timelimit);

	synchronized(mdl) {
	    while (!mdl.finished) {
		try {
		    mdl.wait();
		} catch (InterruptedException ex) {
		}
	    }
	}

	if (mdl.bestSolution.score > -Double.MAX_VALUE)
	    return mdl.bestSolution;
	return null;
    }

    static class TestCase
    {
	String ciphertext;
	String plaintext;
	int    map[];

	int     ntrials;
	int     nsolves;
	double  time;
	double  time2;

	double  bestscore = -Double.MAX_VALUE;
    }

    public static void mainEx(String args[]) throws IOException
    {
	Dict dict;

	try {
	    dict = new Dict("english-standard.dat");
	} catch (IOException ex) {
	    System.out.println("Couldn't open dictionary: "+ex);
	    return;
	}

	BufferedWriter outs = new BufferedWriter(new FileWriter("/tmp/gabenchdata.txt"));
	BufferedReader ins = new BufferedReader(new FileReader(new File(args[0])));

	String line;
	int puzzleNumber = 0;

	ArrayList<TestCase> testcases = new ArrayList<TestCase>();

	while ((line = ins.readLine())!=null) {
	    if (line.startsWith("#"))
		continue;

	    if (line.startsWith("**")) {
		TestCase tc = new TestCase();
		tc.ciphertext = ins.readLine();
		tc.plaintext = ins.readLine();
		tc.map = new int[26];

		for (int i = 0; i < Math.min(tc.ciphertext.length(), tc.plaintext.length()); i++) {
		    char cc = tc.ciphertext.charAt(i);
		    char pc = tc.plaintext.charAt(i);

		    if (Character.isLetter(cc)) {
			cc = Character.toUpperCase(cc);
			pc = Character.toUpperCase(pc);
			tc.map[cc-'A'] = pc-'A';
		    }
		}

		testcases.add(tc);
	    }
	}

	int maxtestcases = Math.min(100, testcases.size());

	for (int trial = 0; trial < 1000; trial++) {

	    for (int tcidx = 0; tcidx < maxtestcases; tcidx++) {
		TestCase tc = testcases.get(tcidx);

		Puzzle puzzle = new Puzzle(dict, tc.ciphertext, "", false, true);
		GAProblem problem = new GAProblem(puzzle);
		Solver solver = new GeneticSolver(problem);
		int identitymap[] = new int[dict.getLanguage().letterCount()+1];

		double scoreThresh = problem.computeScore(tc.map) / puzzle.numCiphertextLetters;

		System.out.println("--------------------------------------");
		System.out.println("puzzle: "+tcidx+"   round: "+trial);
		System.out.println(tc.ciphertext);
		System.out.printf("%10f (%8s  ) %s\n", scoreThresh, "truth", tc.plaintext);

		long startTime = System.currentTimeMillis();
		Solution sol = solveWait(puzzle, solver, tc, 30, scoreThresh);
		long endTime = System.currentTimeMillis();

		double dt = (endTime-startTime)/1000.0;

		System.out.printf("%10f (%8.3f s) %s\n", sol.score, dt, sol.plaintext);
		tc.ntrials++;
		if (sol.solved)
		    tc.nsolves++;
		tc.time += sol.time;
		tc.time2 += sol.time*sol.time;
		tc.bestscore = Math.max(tc.bestscore, sol.score);

		double avgtime = tc.time / tc.ntrials;
		double stdtime = Math.sqrt(tc.time2/tc.ntrials - avgtime*avgtime);

		outs.write(String.format("%5d %3d %15f %15f %15f %15f %15f\n", tcidx, sol.solved ? 1 : 0, sol.time, sol.score, avgtime, stdtime, tc.nsolves / ((double) tc.ntrials)));
		outs.flush();
	    }

	    int ntrials = 0, nsolves = 0;
	    double totalTime = 0;

	    for (int tcidx = 0; tcidx < maxtestcases; tcidx++) {
		TestCase tc = testcases.get(tcidx);

		ntrials += tc.ntrials;
		nsolves += tc.nsolves;
		totalTime += tc.time;
	    }

	    outs.write(String.format("%5d %15f %15f\n", -1, (((double) nsolves)/ntrials), totalTime/ntrials));
	    outs.flush();
	}
    }
}
