package decrypto;

import java.io.*;
import decrypto.dictattack.*;
import decrypto.genetic.*;

public class GATest
{
    public static void main(String args[])
    {
	try {
	    mainEx(args);
	} catch (IOException ex) {
	    System.out.println("ex: "+ex);
	}
    }

    static class Solution
    {
	double score = -Double.MAX_VALUE;
	String plaintext;
    }

    static class MyDecryptoListener implements SolverListener
    {
	Solution bestSolution = new Solution();
	boolean finished = false;
	Puzzle puzzle;

	public void solverMessage(String s)
	{
	}

	public void solverFinished(double elapsedTime)
	{
	    synchronized(this) {
		finished = true;
		notifyAll();
	    }
	}

	public void solverSolution(MapSet mapset, int spaces[], boolean fullSolution, double score)
	{
	    if (score < bestSolution.score)
		return;

	    bestSolution.score = score;
	    bestSolution.plaintext = puzzle.lang.translateString(puzzle.cipherText, mapset, spaces);
	}

	public void solverProgress(double progress)
	{
	}
    }

    static Solution solveWait(Puzzle puzzle, Solver solver, double timelimit)
    {
	MyDecryptoListener mdl = new MyDecryptoListener();
	mdl.puzzle = puzzle;
	solver.addListener(mdl);
	solver.start(timelimit);

	synchronized(mdl) {
	    while (!mdl.finished) {
		try {
		    mdl.wait();
		} catch (InterruptedException ex) {
		}
	    }
	}

	if (mdl.bestSolution.score > -Double.MAX_VALUE)
	    return mdl.bestSolution;
	return null;
    }
     
    public static void mainEx(String args[]) throws IOException
    {
	Dict dict;

	try {
	    dict = new Dict("english-standard.dat");
	} catch (IOException ex) {
	    System.out.println("Couldn't open dictionary: "+ex);
	    return;
	}

	BufferedReader ins = new BufferedReader(new FileReader(new File(args[0])));

	String line;
	int puzzleNumber = 0;

	while ((line = ins.readLine())!=null) {
	    if (line.startsWith("#"))
		continue;
	    if (line.length()<10)
		continue;

	    int offset;
	    for (offset = 0; offset < line.length(); offset++) {
		char c = line.charAt(offset);
		if (Character.isLetter(c) || c=='\"')
		    break;
	    }

	    line = line.substring(offset);

	    System.out.printf("**%d**\n%s\n", puzzleNumber, line);

	    Puzzle puzzle = new Puzzle(dict, line, "", true, true);
	    Solver solver = new DictionaryAttackSolver(puzzle);
	    Solution sol = solveWait(puzzle, solver, 15);

	    System.out.printf("%s\n\n", sol.plaintext);

	    puzzleNumber++;
	}
    }
}
