package decrypto.dictattack;

import decrypto.*;

import java.util.*;

/** LazyPlanner selects the word or letter that has the least
 * fan-out. This is generally a win when the set intersection
 * algorithm is being used to prune the candidate list at each nodes.
 **/
public class LazyPlanner implements Planner
{
    Puzzle puzzle;

    public LazyPlanner(Puzzle p)
    {
	this.puzzle = p;
    }

    public int getAction(boolean solvedWords[], MapSet map)
    {
	// small scores are good.
	double bestscore = Double.MAX_VALUE;
	int bestword = -1;

	for (int i = 0; i < solvedWords.length; i++) {
	    if (solvedWords[i])
		continue;

	    Word w = puzzle.words.get(i);
	    if (!w.enabled)
		continue;

	    double thisscore = (w.candidates.length - w.firstcand);

	    // no candidates left for this word
	    if (thisscore == 0)
		return -2;

	    if (thisscore < bestscore) {
		bestscore = thisscore;
		bestword = i;
	    }
	}

	if (DecryptoParameters.enableSingleLetters && bestscore > puzzle.lang.letterCount()) {

	    // find next untried letter
	    for (int i = 0; i < puzzle.historder.length; i++)  {

		// skip ciphertext letters that we already have a mapping for.
		if (map.isUniquelyMapped(puzzle.historder[i])) // || !map.hasMappings((byte) i))
		    continue;

		byte letter = puzzle.historder[i];

		if (puzzle.histogram[letter] == 0)
		    continue;
		
		return letter | LETTER_FLAG;
	    }
	}

	return bestword | WORD_FLAG;
    }
}
