package decrypto.gui;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.awt.datatransfer.*;
import javax.swing.text.*;

import decrypto.*;
import decrypto.util.*;

public class DecryptoDictGUI
{
    public static final long serialVersionUID=1001;
    JMenu  menu;
							
    JFrame frame = new JFrame("Decrypto Dictionary Editor");

    ArrayList<Word> words = new ArrayList<Word>();
    ArrayList<Word> filteredWords = new ArrayList<Word>();

    WordTableModel tableModel = new WordTableModel();
    JTable wordTable = new JTable(tableModel);
    JLabel statusLabel = new JLabel("No words loaded");
    JLabel progressLabel = new JLabel("");

    JFileChooser fileChooser = new JFileChooser(System.getProperty("user.dir"));
    ExtensionFileFilter txtFilter = new ExtensionFileFilter("txt", "Text files");
    ExtensionFileFilter datFilter = new ExtensionFileFilter("dat", "Dictionary files");
    ExtensionFileFilter txtDatFilter = new ExtensionFileFilter(new String[] {"txt", "dat"}, "Text files and dictionary files");

    ParameterGUI pg = new ParameterGUI();
    Language lang;
    
    JTextField filterField = new JTextField(15);

    public static void main(String args[])
    {
	DecryptoDictGUI d = new DecryptoDictGUI();
    }

    public DecryptoDictGUI()
    {
	pg.addString("langclass","Language Class","decrypto.EnglishLanguage");

	lang = Dict.getLanguage(pg.gs("langclass"));

	Font fnt = new Font("Monospaced", Font.PLAIN, statusLabel.getFont().getSize());
	wordTable.setFont(fnt);
	wordTable.addKeyListener(new MyTableKeyListener());

	JMenuBar menuBar = new JMenuBar();
	frame.setJMenuBar(menuBar);
 	JMenu fileMenu = new JMenu("File");
	menuBar.add(fileMenu);

	fileMenu.add(makeMenuItem("New", KeyEvent.VK_N, 
				  null,
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  clear();
				      }}));


	fileMenu.add(makeMenuItem("Import words...", KeyEvent.VK_I, 
				  null,
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  importWords();
				      }}));

	fileMenu.add(makeMenuItem("Save dictionary...", KeyEvent.VK_S, 
				  null,
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  saveDictionary();
				      }}));
	fileMenu.add(makeMenuItem("Save word list...", 0,
				  null,
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  saveWordList();
				      }}));


	fileMenu.addSeparator();
	fileMenu.add(makeMenuItem("Quit", KeyEvent.VK_Q, 
				  KeyStroke.getKeyStroke(KeyEvent.VK_Q, ActionEvent.CTRL_MASK),
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  frame.dispose();
				      }}));

	JMenu sortMenu = new JMenu("Sort");
	menuBar.add(sortMenu);
	sortMenu.add(makeMenuItem("Alphabetical Order", KeyEvent.VK_A,
				  null,
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  Collections.sort(words, new Word.TextComparator());
					  update();
				      }}));
	sortMenu.add(makeMenuItem("Pattern Order", KeyEvent.VK_D, 
				  null,
				  new ActionListener() {
				      public void actionPerformed(ActionEvent e) {
					  Collections.sort(words, new Word.PatternComparator());
					  update();
				      }}));

	JMenu toolsMenu = new JMenu("Tools");
	menuBar.add(toolsMenu);
	toolsMenu.add(makeMenuItem("Add words manually...", KeyEvent.VK_A,
				   null,
				   new ActionListener() {
				       public void actionPerformed(ActionEvent e) {
					   addWordsManually();
				       }}));

	frame.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    frame.dispose();
		}});

	frame.setLayout(new BorderLayout());
	frame.add(new JScrollPane(wordTable), BorderLayout.CENTER);

	JPanel panel = new JPanel(new BorderLayout());
	panel.add(statusLabel, BorderLayout.WEST);
	JPanel panel2 = new JPanel(new BorderLayout());
	panel2.add(DecryptoGUI.makeLabel("   Filter: ", 'f'), BorderLayout.WEST);
	panel2.add(filterField, BorderLayout.CENTER);
	JButton clearFilterButton = new JButton("Clear");
	clearFilterButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    filterField.setText("");
		    update();
		}});
	panel2.add(clearFilterButton, BorderLayout.EAST);

	filterField.setFocusAccelerator('f');

	filterField.getDocument().addDocumentListener(new DocumentListener() {
		public void changedUpdate(DocumentEvent e) {
		    update();
		}

		public void insertUpdate(DocumentEvent e) {
		    update();
		}

		public void removeUpdate(DocumentEvent e) {
		    update();
		}
	    });
		    
	panel.add(panel2, BorderLayout.CENTER);
	panel.add(progressLabel, BorderLayout.SOUTH);

	frame.add(panel, BorderLayout.SOUTH);
	frame.add(pg.getPanel(), BorderLayout.NORTH);

	TableColumnModel tcm = wordTable.getColumnModel();
        tcm.getColumn(0).setMaxWidth(100);

	Image icon = Toolkit.getDefaultToolkit().getImage(getClass().getResource("/decrypto.png"));
	frame.setIconImage(icon);

	fileChooser.setFileFilter(txtFilter);
	fileChooser.setFileFilter(datFilter);
	fileChooser.setFileFilter(txtDatFilter);

	frame.setSize(400,600);
	frame.setVisible(true);
	update();
    }

    class WordEditor
    {
	JFrame wframe;
	JTextArea textArea = new JTextArea();
	JButton addButton = new JButton("Add");
	JButton cancelButton = new JButton("Cancel");

	WordEditor()
	{
	    wframe = new JFrame("Add words");
	    wframe.setLayout(new BorderLayout());
	    wframe.add(new JLabel("Type in new words, one per line."), BorderLayout.NORTH);

	    JPanel p = new JPanel(new GridLayout(1,2));
	    p.add(cancelButton);
	    p.add(addButton);
	    wframe.add(textArea, BorderLayout.CENTER);
	    wframe.add(p, BorderLayout.SOUTH);
	    wframe.setSize(400,400);
	    wframe.setVisible(true);

	    wframe.addWindowListener(new WindowAdapter() {
		    public void windowClosing(WindowEvent e) {
			wframe.dispose();
		    }});

	    cancelButton.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			wframe.dispose();
		    }});

	    addButton.addActionListener(new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			if (lang == null)
			    lang = new EnglishLanguage();

			String t[] = textArea.getText().split("\\s+");
			for (int i = 0; i < t.length; i++)
			    {
				if (t[i].charAt(0)=='#')
				    continue;

				byte[] bs = lang.stringToBytes(t[i]);
				if (bs == null)
				    continue;
				
				words.add(new Word(lang.bytesToString(bs), 0, bs));
			    }
			words = Dict.sortAndPruneWords(words);
			update();
			wframe.dispose();
		    }});
	}
    }

    void addWordsManually()
    {
	new WordEditor();
    }

    void update()
    {
	// fix up the search string; remove punctuation, etc.
	String s = filterField.getText().toUpperCase();
	s = lang.bytesToString(lang.stringToBytes(s));

	int currentPosition = wordTable.getSelectedRow();
	if (currentPosition < 0 || currentPosition >= words.size())
	    currentPosition = 0;


	filteredWords.clear();

	for (int i = 0; i < words.size(); i++) {
	    Word w = words.get(i);

	    if (w.text.startsWith(s)) {
		filteredWords.add(w);
	    }
	}

	statusLabel.setText("Size: "+words.size()+" words");

	if (s.length()==0) {
	    progressLabel.setText("");
	} else {
	    progressLabel.setText(String.format("Filter: %d of %d words begin with %s", 
						filteredWords.size(), words.size(), s));
	}

	tableModel.fireTableDataChanged();
    }

    void clear()
    {
	words = new ArrayList<Word>();
	update();
    }

    void importWords()
    {
	fileChooser.setFileFilter(txtDatFilter);

	int res = fileChooser.showOpenDialog(frame);

	if (res != JFileChooser.APPROVE_OPTION)
	    return;

	String path = fileChooser.getSelectedFile().getAbsolutePath();
	if (path.toLowerCase().endsWith(".dat"))
	    loadFromDictionary(path);
	else
	    loadFromWordList(path);

	progressLabel.setText("Words imported.");
    }

    void loadFromDictionary()
    {
	fileChooser.setFileFilter(datFilter);

	int res = fileChooser.showOpenDialog(frame);

	if (res != JFileChooser.APPROVE_OPTION)
	    return;

	String path = fileChooser.getSelectedFile().getAbsolutePath();
	loadFromDictionary(path);
    }

    void loadFromDictionary(String path)
    {
	try {
	    Dict d = new Dict(path);
	    lang = d.getLanguage();
	    
	    ArrayList<byte[]> pats = d.readAll();
	    for (int i = 0; i < pats.size(); i++)
		{
		    words.add(new Word(lang.bytesToString(pats.get(i)), 0, pats.get(i)));
		}
	} catch (IOException ex) {
	    JOptionPane.showMessageDialog(null, "Couldn't read dictionary: "+ex, "Error", JOptionPane.ERROR_MESSAGE);
	}

	words = Dict.sortAndPruneWords(words);

	update();
    }

    void loadFromWordList()
    {
	fileChooser.setFileFilter(txtFilter);

	int res = fileChooser.showOpenDialog(frame);

	if (res != JFileChooser.APPROVE_OPTION)
	    return;

	String path = fileChooser.getSelectedFile().getAbsolutePath();
	loadFromWordList(path);
    }

    void loadFromWordList(String path)
    {
	try {
	    FileReader fr = new FileReader(path);
	    BufferedReader ins = new BufferedReader(fr);
	    
	    lang = Dict.getLanguage(pg.gs("langclass"));
	    String line;
	    while ((line = ins.readLine())!=null) 
		{
		    line = line.trim();
		    if (line.length()==0)
			continue;

		    // skip comments
		    if (line.charAt(0)=='#')
			continue;

		    byte[] bs = lang.stringToBytes(line);
		    if (bs == null)
			continue;
		    
		    words.add(new Word(lang.bytesToString(bs), 0, bs));
		}
	} catch (IOException ex) {
	    JOptionPane.showMessageDialog(null, "Couldn't read word list: "+ex, "Error", JOptionPane.ERROR_MESSAGE);
	}

	words = Dict.sortAndPruneWords(words);
	update();
    }

    void saveWordList()
    {
	fileChooser.setFileFilter(txtFilter);

	// Try coaxing the default filename to end in ".txt" instead
	// of .dat, per one of the open bugs. Sadly, this doesn't
	// work. (f==null).
	if (false) {
	    File f = fileChooser.getSelectedFile();
	    if (f != null) {
		String path = f.getAbsolutePath();
		if (path.endsWith(".dat")) {
		    path = path.substring(0, path.length()-4)+".txt";
		    System.out.println(path);
		    fileChooser.setSelectedFile(new File(path));
		}
	    }
	}

	int res = fileChooser.showSaveDialog(frame);

	if (res != JFileChooser.APPROVE_OPTION)
	    return;

	String path = fileChooser.getSelectedFile().getAbsolutePath();
	if (path.endsWith(".dat")) {
	    path = path.substring(0, path.length()-4)+".txt";
	    System.out.println(path);
	    fileChooser.setSelectedFile(new File(path));
	}

	try {
	    BufferedWriter outs = new BufferedWriter(new FileWriter(path));
	    for (int i = 0; i < words.size(); i++)
		{
		    outs.write(words.get(i).text);
		    outs.newLine();
		}
	    outs.close();
	} catch (IOException ex) {
	    JOptionPane.showMessageDialog(null, "Couldn't write word list: "+ex, "Error", JOptionPane.ERROR_MESSAGE);
	}
    }
    

    void saveDictionary()
    {
	fileChooser.setFileFilter(datFilter);

	int res = fileChooser.showSaveDialog(frame);

	if (res != JFileChooser.APPROVE_OPTION)
	    return;

	String path = fileChooser.getSelectedFile().getAbsolutePath();

	Dict dict = new Dict();
	dict.properties.put("langclass", pg.gs("langclass"));
	dict.properties.put("version", "2");
	try {
	    dict.write(path, words);
	} catch (IOException ex) {
	    JOptionPane.showMessageDialog(null, "Couldn't write dictionary: "+ex, "Error", JOptionPane.ERROR_MESSAGE);
	}
    }

    protected JMenuItem makeMenuItem(String name, int mnemonic, KeyStroke ks, ActionListener listener)
    {
	JMenuItem jmi = new JMenuItem(name, mnemonic);
	jmi.addActionListener(listener);
	jmi.setAccelerator(ks);
	return jmi;
    }

    class WordTableModel extends AbstractTableModel
    {
	public static final long serialVersionUID = 1001;

	public int getRowCount()
	{
	    return filteredWords.size();
	}

	public int getColumnCount()
	{
	    return 2;
	}

	public Object getValueAt(int row, int col)
	{
	    Word w = filteredWords.get(row);
	    switch (col)
		{
		case 0:
		    return ""+(row+1);
		case 1:
		    return w.text;
		}
	    return "";
	}

	public String getColumnName(int col)
	{
	    switch (col)
		{
		case 0:
		    return "Row";
		case 1:
		    return "Word";
		}
	    return "?";
	}
    }

    class MyTableKeyListener extends KeyAdapter
    {
	public void keyPressed(KeyEvent e)
	{
	    if (e.getKeyCode() == KeyEvent.VK_DELETE ||
		e.getKeyCode() == KeyEvent.VK_BACK_SPACE) {

		int[] idxs = wordTable.getSelectedRows();
		HashSet<Word> deletewords = new HashSet<Word>();
		for (int j = 0; j < idxs.length; j++)
		    deletewords.add(filteredWords.get(idxs[j]));

		ArrayList<Word> newwords = new ArrayList<Word>();

		for (int i = 0; i < words.size(); i++) {
		    Word w = words.get(i);
		    if (!deletewords.contains(w))
			newwords.add(w);
		}
		
		words = newwords;
		update();
	    }
	}
    }
}


