package decrypto.statistics;

import java.io.*;
import java.util.zip.*;

public class BuildStatistics3
{
    // 0-25: A-Z
    // 26: space
    // 27: sum of 0-26
    static final int SZ = 28;
    static final int SUM = SZ-1;
    static final int SPACE = SZ-2;

    double  counts[][][]; 
    int     history[];       // [order]

    int     totalChars;

    public static void main(String args[])
    {
	try {
	    main_ex(args);
	} catch (IOException ex) {
	    System.out.println("ex: "+ex);
	}
    }
    
    public static void main_ex(String args[]) throws IOException
    {
	BuildStatistics3 bs1 = new BuildStatistics3(args, true, "english3");
	BuildStatistics3 bs2 = new BuildStatistics3(args, false, "english3ns");
    }

    public BuildStatistics3(String args[], boolean spaces, String classname) throws IOException
    {
	counts = new double[SZ][SZ][SZ];
	history = new int[3];

	for (int i = 0; i < history.length; i++)
	    history[i] = SPACE;
	for (int i = 0; i < args.length; i++) {
	    FileReader fr = new FileReader(args[i]);
	    BufferedReader ins = new BufferedReader(fr);

	    int c;
	    while ((c = ins.read())>0) {
		if (Character.isLetter(c))
		    handleCharacter(Character.toUpperCase(c)-'A');
		else if (spaces)
		    handleCharacter(SPACE);
	    }
	}

	System.out.println("Total characters: " + totalChars);

	normalize();
	String outpath = "/tmp/"+classname+".java";
	System.out.println("Writing to: "+outpath);
	//	emit(classname, outpath);
	emitBinary("/tmp/"+classname+".sta");
    }

    public void handleCharacter(int v)
    {
	// skip multiple spaces in a row
	if (v==SPACE && history[history.length-1]==SPACE)
	    return;

	if (v>=SZ) 
	    return;

	for (int i = 1; i < history.length; i++) 
	    history[i-1] = history[i];
	history[history.length-1] = v;

	counts[history[0]][history[1]][history[2]]++;
	counts[history[0]][history[1]][SUM]++;
	totalChars++;

	// Whenever we see the first or last character of a word, we
	// also update SPACE-SPACE-LETTER and LETTER-SPACE-SPACE. This
	// helps us compute better statistics on the extreme ends of a
	// puzzle, where we do not know the preceding/following word.
	if (history[1]==SPACE) {
	    counts[history[0]][SPACE][SPACE]++;
	    counts[history[0]][SPACE][SUM]++;

	    counts[SPACE][SPACE][history[2]]++;
	    counts[SPACE][SPACE][SUM]++;
	}
    }

    public void normalize()
    {
	for (int i0 = 0; i0 < counts.length; i0++) {
	    for (int i1 = 0; i1 < counts[i0].length; i1++) {
		for (int i2 = 0; i2 < counts[i0][i1].length; i2++) {
			double v = counts[i0][i1][i2];

			double frac = (counts[i0][i1][i2]+1) / (counts[i0][i1][SUM] + SZ);
			counts[i0][i1][i2] = Math.log(frac);
		}
	    }
	}
    }

    public static float[][][] readBinary(String path) throws IOException
    {
	long starttime = System.currentTimeMillis();
	System.out.print("Opening "+path+" ... ");
	System.out.flush();

	DataInputStream ins = new DataInputStream(new BufferedInputStream(new GZIPInputStream(new BufferedInputStream(new FileInputStream(path)))));
	int dim = ins.readInt();
	int sz = ins.readInt();

	assert(dim==3);

	float data[][][] = new float[sz][sz][sz];
	for (int i = 0; i < SZ; i++) {
	    for (int j = 0; j < SZ; j++) {
		for (int k = 0; k < SZ; k++) {
		    data[i][j][k] = ins.readFloat();
		}
	    }
	}

	System.out.printf("done (%.3f s)\n", (System.currentTimeMillis()-starttime)/1000.0);

	return data;
    }

    public void emitBinary(String path) throws IOException
    {
	DataOutputStream outs = new DataOutputStream(new GZIPOutputStream(new FileOutputStream(path)));
	outs.writeInt(3);  // order of model
	outs.writeInt(SZ); // dimensionality
	
	for (int i = 0; i < SZ; i++) {
	    for (int j = 0; j < SZ; j++) {
		for (int k = 0; k < SZ; k++) {
		    outs.writeFloat((float) counts[i][j][k]);
		}
	    }
	}
	outs.close();
    }

    public void emit(String classname, String path) throws IOException 
    {
	PrintStream outs = new PrintStream(new BufferedOutputStream(new FileOutputStream(path)));

	outs.printf("package decrypto.statistics;\n\n");
	outs.printf("public class "+classname+"\n{\n");

	String tab = "\t";

	for (int i = 0; i < SZ; i++) {

	    outs.printf("static final float[][] trigrams%d() { \n%sreturn new float[][] \n", i, tab);
		outs.printf("%s{\n", tab);

		for (int j = 0; j < 28; j++) {
		    outs.printf("%s%s// %c%c\n",tab,tab, i+'A', j+'A');

		    outs.printf("%s%s{  ",tab,tab);
		    for (int k = 0; k < 28; k++)
			{
			    float v = (float) counts[i][j][k];
			    float lv = v;

			    outs.printf("%.4ff", lv);
			    if (k<27)
				outs.printf(", ");
			    else
				outs.printf(" ");
			    if ((k&7)==7)
				outs.printf("\n%s%s%s",tab,tab,tab);
			}
		    outs.printf("}");
		    if (j<27)
			outs.printf(",\n");
		    else
			outs.printf("\n");
		}
		outs.printf("%s};\n}\n",tab);
	}
	
	
	outs.printf("public static final float trigrams[][][] = new float[][][] {\n");

	for (int i = 0; i < SZ; i++) {
	    outs.printf("trigrams%d()", i);
	    if (i<27)
		outs.printf(", ");
	    else
		outs.printf(" ");
	}

	outs.printf("};\n");
	
	outs.printf("}\n");
	outs.flush();
    }
}
    
