package decrypto.statistics;

import java.io.*;
import java.util.zip.*;

public class BuildStatistics4
{
    // 0-25: A-Z
    // 26: space
    // 27: sum of 0-26/expectation
    static final int SZ = 28;
    static final int SPACE = SZ-2;
    static final int SUM = SZ-1;  // used for expectation when we write it out....
    static final int EXPECT = SZ - 1;

    double  counts[][][][]; 
    int     history[];       // [order]

    double  counts3[][][];
    double  counts2[][];
    double  counts1[];

    int     totalChars;

    public static void main(String args[])
    {
	try {
	    main_ex(args);
	} catch (IOException ex) {
	    System.out.println("ex: "+ex);
	}
    }
    
    public static void main_ex(String args[]) throws IOException
    {
	BuildStatistics4 bs1 = new BuildStatistics4(args, true, "english4");
	BuildStatistics4 bs2 = new BuildStatistics4(args, false, "english4ns");
    }


    public BuildStatistics4(String args[], boolean spaces, String classname) throws IOException
    {
	counts = new double[SZ][SZ][SZ][SZ];
	history = new int[4];
	counts3 = new double[SZ][SZ][SZ];
	counts2 = new double[SZ][SZ];
	counts1 = new double[SZ];

	for (int i = 0; i < history.length; i++)
	    history[i] = SPACE;

	for (int i = 0; i < args.length; i++) {
	    FileReader fr = new FileReader(args[i]);
	    BufferedReader ins = new BufferedReader(fr);

	    int c;
	    while ((c = ins.read())>0) {
		if (Character.isLetter(c))
		    handleCharacter(Character.toUpperCase(c)-'A');
		else if (spaces)
		    handleCharacter(SPACE);
	    }
	}

	System.out.println("Total characters: "+totalChars);

	normalize();
	String outpath = "/tmp/"+classname+".java";
	System.out.println("Writing to: "+outpath);
	//emit(classname, outpath);
	emitBinary("/tmp/"+classname+".sta");
    }

    public void handleCharacter(int v)
    {
	// skip multiple spaces in a row
	if (v==SPACE && history[history.length-1]==SPACE)
	    return;

	if (v>=SZ) 
	    return;

	for (int i = 1; i < history.length; i++) 
	    history[i-1] = history[i];
	history[history.length-1] = v;

	counts[history[0]][history[1]][history[2]][history[3]]++;
	counts[history[0]][history[1]][history[2]][SUM]++;

	counts3[history[1]][history[2]][history[3]]++;
	counts3[history[1]][history[2]][SUM]++;

	counts2[history[2]][history[3]]++;
	counts2[history[2]][SUM]++;

	counts1[history[3]]++;
	counts1[SUM]++;

	totalChars++;
    }

    public double expectedInformation(double cs[])
    {
	double expectation = 0;

	for (int i = 0; i <= SPACE; i++) {
	    double prob = (cs[i]+1)/(cs[SUM]+27);
	    double information = Math.log(1.0/prob);
	    expectation += prob*information;
	}

	return expectation;
    }

    public void normalize()
    {
	// some count tables may have very small numbers of
	// occurences. That's a serious problem for the genetic
	// algorithm, which may start with highly unusual initial
	// states. It's important that it be able to find a gradient.
	// Hence, our strategy here is to replace the 4-gram with a
	// 3-gram table if the 3-gram table would do better. (And the
	// 3-gram table with a 2-gram table, etc.) We pick the table
	// that has the least amount of expected information. (i.e.,
	// the one that we know the most about a priori)

	for (int i0 = 0; i0 < counts.length; i0++) {
	    for (int i1 = 0; i1 < counts[i0].length; i1++) {
		for (int i2 = 0; i2 < counts[i0][i1].length; i2++) {
		    double v[] = counts[i0][i1][i2];
		    //   System.out.printf("%c%c%c  %10f %10f %10f %10f\n", i0+'A', i1+'A', i2+'A', information4, information3, information2, information1);

		    double samples = counts[i0][i1][i2][SUM];
		    double thresh = 5;
		    double offset = 0;

		    if (samples < thresh) {
			v = counts3[i1][i2];
			samples = counts3[i1][i2][SUM];
			offset = 10*counts3[i1][i2][SUM];
		    }

		    if (samples < thresh) {
			v = counts2[i2];
			samples = counts2[i2][SUM];
			offset = 100 * counts2[i2][SUM];
		    }

		    /*
		    if (samples < thresh) {
			v = counts1;
			samples = counts1[SUM];
		    }
		    */

		    /*
		    double information4 = expectedInformation(counts[i0][i1][i2]);
		    double information3 = expectedInformation(counts3[i1][i2]);
		    double information2 = expectedInformation(counts2[i2]);
		    double information1 = expectedInformation(counts1);

		    if (information3 < information4) {
			v = counts3[i1][i2];
			information4 = information3;
		    }

		    if (information2 < information4) {
			v = counts2[i2];
			information4 = information2;
		    }


		    if (information1 < information4) {
			v = counts1;
			information4 = information1;
		    }
		    */

		    double sum = 0;
		    for (int i = 0; i <= SPACE; i++) {
			counts[i0][i1][i2][i] = v[i] + offset;
			sum += counts[i0][i1][i2][i];
		    }
		    counts[i0][i1][i2][SUM] = sum;
		}
	    }
	}
	
	// divide each count by the total number in that category
	for (int i0 = 0; i0 < counts.length; i0++) {
	    for (int i1 = 0; i1 < counts[i0].length; i1++) {
		for (int i2 = 0; i2 < counts[i0][i1].length; i2++) {
		    for (int i3 = 0; i3 < counts[i0][i1][i2].length; i3++) {
			double v = counts[i0][i1][i2][i3];

			double frac = (counts[i0][i1][i2][i3]+1) / (counts[i0][i1][i2][SUM] + SZ);
			double logprob = Math.log(frac);
			counts[i0][i1][i2][i3] = logprob;
		    }
		}
	    }
	}

	// We don't always know the mapping for a particular letter.
	// Elsewhere, we might call these (erroneously) marginals, but
	// they're really expected values.  e.g., if I wnat the n-gram
	// probability for the character sequence AB?D, I can't
	// compute it because I don't know what ? is. However, I can
	// compute the expected value. We also have to deal with
	// patterns like ??AB.
	for (int i0 = 0; i0 < counts.length; i0++) {
	    for (int i1 = 0; i1 < counts[i0].length; i1++) {
		for (int i2 = 0; i2 < counts[i0][i1].length; i2++) {
		    for (int i3 = 0; i3 < counts[i0][i1][i2].length; i3++) {

			// no expectations to compute here!
			if (i0 != EXPECT && i1 != EXPECT && i2 != EXPECT && i3 != EXPECT)
			    continue;
			
			// now, for each unknown variable (where i ==
			// EXPECT), loop over all values.
			int j0a = i0, j0b = i0; // if i0 isn't EXPECT, we don't loop over this var.
			if (i0 == EXPECT) {
			    j0a = 0; j0b = SPACE; // we do need to loop...
			}

			int j1a = i1, j1b = i1;
			if (i1 == EXPECT) {
			    j1a = 0; j1b = SPACE;
			}

			int j2a = i2, j2b = i2;
			if (i2 == EXPECT) {
			    j2a = 0; j2b = SPACE;
			}

			int j3a = i3, j3b = i3;
			if (i3 == EXPECT) {
			    j3a = 0; j3b = SPACE;
			}

			double probsum = 0;
			double expectationsum = 0;

			for (int j0 = j0a; j0 <= j0b; j0++) {
			    for (int j1 = j1a; j1 <= j1b; j1++) {
				for (int j2 = j2a; j2 <= j2b; j2++) {
				    for (int j3 = j3a; j3 <= j3b; j3++) {

					double prob = Math.exp(counts[j0][j1][j2][j3]);
					probsum += prob;
					expectationsum += counts[j0][j1][j2][j3]*prob;
				    }
				}
			    }
			}

			counts[i0][i1][i2][i3] = expectationsum / probsum;

			//System.out.printf("%c%c%c%c : %.5f\n", i0+'A', i1+'A', i2+'A', i3+'A', counts[i0][i1][i2][i3]);
		    }
		}
	    }
	}		    
    }

    public static float[][][][] readBinary(String path) throws IOException
    {
	long starttime = System.currentTimeMillis();
	//	System.out.print("Opening "+path+" ... ");
	System.out.flush();

	DataInputStream ins = new DataInputStream(new BufferedInputStream(new GZIPInputStream(new BufferedInputStream(new FileInputStream(path)))));

	int dim = ins.readInt();
	int sz = ins.readInt();

	assert(dim==4);

	float data[][][][] = new float[sz][sz][sz][sz];
	for (int i = 0; i < SZ; i++) {
	    for (int j = 0; j < SZ; j++) {
		for (int k = 0; k < SZ; k++) {
		    for (int w = 0; w < SZ; w++) {
			data[i][j][k][w] = ins.readFloat();
		    }
		}
	    }
	}

	//	System.out.printf("done (%.3f s)\n", (System.currentTimeMillis()-starttime)/1000.0);
	return data;
    }

    public void emitBinary(String path) throws IOException
    {
	DataOutputStream outs = new DataOutputStream(new GZIPOutputStream(new FileOutputStream(path)));
	
	outs.writeInt(4);  // order of model
	outs.writeInt(SZ); // dimensionality
	
	for (int i = 0; i < SZ; i++) {
	    for (int j = 0; j < SZ; j++) {
		for (int k = 0; k < SZ; k++) {
		    for (int w = 0; w < SZ; w++) {
			outs.writeFloat((float) counts[i][j][k][w]);
		    }
		}
	    }
	}
	outs.close();
    }

    public void emit(String classname, String path) throws IOException 
    {
	PrintStream outs = new PrintStream(new BufferedOutputStream(new FileOutputStream(path)));

	outs.printf("package decrypto.statistics;\n\n");
	outs.printf("public class "+classname+"\n{\n");

	String tab = "\t";

	for (int i0 = 0; i0 < SZ; i0++) {
	    for (int i1 = 0; i1 < SZ; i1++) {

		outs.printf("static final float[][] tetragrams%d_%d() { \nreturn new float[][] \n", i0, i1);
		outs.printf("%s{\n", tab);

		for (int j = 0; j < 28; j++) {
		    outs.printf("%s%s// %c%c%c\n",tab,tab, i0+'A', i1+'A',j+'A');

		    outs.printf("%s%s{  ",tab,tab);
		    for (int k = 0; k < 28; k++)
			{
			    float v = (float) counts[i0][i1][j][k];
			    float lv = v;

			    outs.printf("%.4ff", lv);
			    if (k<27)
				outs.printf(", ");
			    else
				outs.printf(" ");
			    if ((k&7)==7)
				outs.printf("\n%s%s%s",tab,tab,tab);
			}
		    outs.printf("}");
		    if (j<27)
			outs.printf(",\n");
		    else
			outs.printf("\n");
		}
		outs.printf("%s};\n}\n",tab);
	    }
	}

	outs.printf("public static final float tetragrams[][][][] = new float[][][][] {\n");

	for (int i0 = 0; i0 < SZ; i0++) {
	    outs.printf("{");
	    for (int i1 = 0; i1 < SZ; i1++) {
		outs.printf("tetragrams%d_%d()", i0, i1);
		if (i1<27)
		    outs.printf(", ");
		else
		    outs.printf(" ");
	    }
	    outs.printf("}");
	    if (i0<27)
		outs.printf(", \n");
	    else
		outs.printf(" \n");
	}

	outs.printf("};\n");
	
	outs.printf("}\n");
	outs.flush();
    }
}
    
