package decrypto.statistics;

import java.io.*;
import java.util.zip.*;
import java.util.*;

public class BuildWordStats
{
    HashMap<String, Word> wordsMap = new HashMap<String, Word>();
    int nwords;
    StringBuffer currentWord = new StringBuffer();

    static final int SPACE = 99999;

    static final class Word implements Comparable<Word>
    {
	String text;
	int    count;

	Word(String text, int count)
	{
	    this.text = text;
	    this.count = count;
	}

	public int compareTo(Word b)
	{
	    return b.count - this.count;
	}
    }

    public static void main(String args[])
    {
	try {
	    main_ex(args);
	} catch (IOException ex) {
	    System.out.println("ex: "+ex);
	}
    }
    
    public static void main_ex(String args[]) throws IOException
    {
	BuildWordStats bws = new BuildWordStats(args);
    }


    public BuildWordStats(String args[]) throws IOException
    {
	for (int i = 0; i < args.length; i++) {
	    FileReader fr = new FileReader(args[i]);
	    BufferedReader ins = new BufferedReader(fr);

	    int c;
	    while ((c = ins.read())>0) {
		if (Character.isLetter(c))
		    handleCharacter(Character.toUpperCase(c));
		else
		    handleCharacter(SPACE);
	    }
	}

	System.out.printf("Num Words: %d\n", nwords);

	System.out.println("Sorting");
	ArrayList<Word> wordList = new ArrayList<Word>();

	for (Word word : wordsMap.values()) {
	    wordList.add(word);
	}

	// Sort with most frequent words first
	Collections.sort(wordList);

	System.out.printf("%d\n", wordList.get(0).count);
	System.out.printf("%d\n", wordList.get(wordList.size()-1).count);

	double cumulativeProb = 0;

	for (int i = 0; i < 10000; i++) {
	    Word word = wordList.get(i);
	    double prob = ((double) word.count)/nwords;
	    cumulativeProb += prob;

	    System.out.printf("%4d %12s (%15f %15f)\n", i, word.text, prob, cumulativeProb);
	}
    }

    void handleCharacter(int c)
    {
	if (c==SPACE) {
	    String text = currentWord.toString();
	    currentWord = new StringBuffer();

	    if (text.length()==0)
		return;

	    Word word = wordsMap.get(text);
	    if (word == null) {
		word = new Word(text, 0);
		wordsMap.put(text, word);
	    }

	    word.count += text.length();
	    nwords     += text.length();
	} else {
	    currentWord.append((char) c);
	}
    }
}
    
