package decrypto.statistics;

import decrypto.*;
import decrypto.genetic.*;

import java.io.*;
import java.util.zip.*;
import java.util.*;

public class GramFreq
{
    // HashMap<patternstring, HashMap<gramstring, Gram>>
    HashMap<String, HashMap<String, Gram>> gramsets = new HashMap<String, HashMap<String,Gram>>();

    // HashMap<gramstring, Gram>
    HashMap<String, Gram> allgrams = new HashMap<String, Gram>();

    int totalCount = 0;
    int order;

    String history = "";

    static final double FRACTOKEEP = 0.6;

    boolean stripspaces = true;

    static final class Gram implements Comparable<Gram>
    {
	String  text;
	int     count;

	Gram(String s)
	{
	    this.text = s;
	}

	public int compareTo(Gram g)
	{
	    return g.count - count;
	}
    }

    public static void main(String args[])
    {
	try {
	    main_ex(args);
	} catch (IOException ex) {
	    System.out.println("ex: "+ex);
	}
    }
    
    public static void main_ex(String args[]) throws IOException
    {
	new GramFreq(5, true, args);
	new GramFreq(5, false, args);
    }

    public GramFreq(int order, boolean stripspaces, String args[]) throws IOException
    {
	this.order = order;
	this.stripspaces = stripspaces;

	for (int i = 0; i < args.length; i++) {
	    FileReader fr = new FileReader(args[i]);
	    BufferedReader ins = new BufferedReader(fr);

	    int c;
	    while ((c = ins.read())>0) {
		if (Character.isLetter(c))
		    handleCharacter(Character.toUpperCase(c));
		else
		    handleCharacter(' ');
	    }
	}

	///////////////////////////////////////////
	// We don't want to report ALL possible NGrams:
	// there are too many. Instead, we report only the most
	// common fraction.

	// First, sort all grams by frequency.
	ArrayList<Gram> allgramslist = new ArrayList<Gram>();
	for (Gram g : allgrams.values())
	    allgramslist.add(g);
	Collections.sort(allgramslist);

	double cumulativeFraction = 0;
	int    countThreshold = 0;
	
	// starting with the most common grams, find what count
	// threshold gives us a cumulative probability of FRACTOKEEP.
	for (int i = 0; i < allgramslist.size(); i++) {
	    Gram g = allgramslist.get(i);
	    cumulativeFraction += ((double) g.count)/totalCount;
	    if (cumulativeFraction > FRACTOKEEP) {
		countThreshold = g.count;
		break;
	    }
	}

	HashMap<String, HashMap<String, Gram>> keepsets = new HashMap<String, HashMap<String, Gram>>();

	//////////////////////////////////////////
	// Now, output the most common grams.
	for (String pattern : gramsets.keySet()) {
	    HashMap<String, Gram> grams = gramsets.get(pattern);

	    // conver the hashmap to a list.
	    ArrayList<Gram> gramlist = new ArrayList<Gram>();
	    for (Gram g : grams.values())
		gramlist.add(g);

	    Collections.sort(gramlist);

	    HashMap<String, Gram> keepgrams = new HashMap<String, Gram>();

	    for (int i = 0; i < gramlist.size(); i++) {
		Gram gram = gramlist.get(i);
		if (gram.count < countThreshold)
		    continue;

		keepgrams.put(gram.text, gram);
		//System.out.printf("%-10s %-10s %10d\n", pattern, gram.text, gram.count);
	    }

	    if (keepgrams.size()>0)
		keepsets.put(pattern, keepgrams);
	}

	gramsets = keepsets;

	emitBinary("/tmp/gramfreq"+order+(stripspaces ? "ns" : "") + ".stb");
    }

    void handleCharacter(int c)
    {
	if (stripspaces && c==' ')
	    return;

	// skip multiple consecutive spaces.
	if (c==' ' && history.length()>0 && history.charAt(history.length()-1)==' ')
	    return;

	if (history.length() == order)
	    history = history.substring(1, order);
	history += (char) c;

	if (history.length() < order)
	    return;

	String pattern = makePattern(history);

	HashMap<String, Gram> grams = gramsets.get(pattern);
	if (grams == null) {
	    grams = new HashMap<String, Gram>();
	    gramsets.put(pattern, grams);
	}

	Gram gram = grams.get(history);
	if (gram == null) {
	    gram = new Gram(history);
	    grams.put(history, gram);
	    allgrams.put(history, gram);
	}
	    
	gram.count++;
	totalCount++;
    }

    static EnglishLanguage lang = new EnglishLanguage();

    public static String makePattern(String s)
    {
	byte map[] = new byte[256];
	StringBuffer pattern = new StringBuffer();
	byte next = 'A';

	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (c==' ') {
		pattern.append(' ');
		continue;
	    }

	    if (map[c]==0) {
		map[c] = next;
		next++;
	    }
	    pattern.append((char) (map[c]));
	}

	return pattern.toString();
    }

    public void emitBinary(String path) throws IOException
    {
	DataOutputStream outs = new DataOutputStream(new GZIPOutputStream(new FileOutputStream(path)));
	
	outs.writeInt(0xaa442a39); // MAGIC
	outs.writeInt(order);      // order of model (length of every gram!)

	outs.writeInt(gramsets.size()); // number of patterns
	for (String pattern : gramsets.keySet()) {
	    HashMap<String, Gram> grams = gramsets.get(pattern);
	    outs.writeInt(grams.size()); // number of grams
	    assert(grams.size()>0);

	    // write the grams
	    for (Gram gram : grams.values()) {
		assert (gram.text.length()==order);
		for (int i = 0; i < order; i++) {
		    outs.write((byte) gram.text.charAt(i));
		}
		outs.writeDouble(gram.count);
	    }
	}

	outs.close();
    }

    public static HashMap<String, Sampler<String>> readBinary(String path) throws IOException
    {
	DataInputStream ins = new DataInputStream(new BufferedInputStream(new GZIPInputStream(new BufferedInputStream(new FileInputStream(path)))));

	int magic = ins.readInt();
	assert(magic == 0xaa442a39);

	int order = ins.readInt();

	HashMap<String, Sampler<String>> map = new HashMap<String, Sampler<String>>();

	int npatterns = ins.readInt();
	for (int pattern = 0; pattern < npatterns; pattern++) {
	    int ngrams = ins.readInt();

	    double proportions[] = new double[ngrams];
	    String strings[] = new String[ngrams];

	    for (int gram = 0; gram < ngrams; gram++) {
		StringBuffer sb = new StringBuffer();

		for (int i = 0; i < order; i++)
		    sb.append((char) ins.read());

		strings[gram] = sb.toString();
		proportions[gram] = ins.readDouble();
	    }

	    map.put(makePattern(strings[0]), new Sampler<String>(proportions, strings));
	}

	return map;
    }
}
