#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>

/* last N characters. Lower indices are more recent. */
char history[2];

/* Array format:
   0:  A
   1:  B
   2:  C
   ...
   24: Y
   25: Z
   26: space
   27: unknown (marginalized)
*/

double count[28][28][28];
double total[28];
double grandtotal;

void update(char c)
{
    // Don't do statistics of multiple whitespace 
	if (history[0]==26 && c==' ')
		return;

//	printf("%c", c);
	if (c>='A' && c<='Z')
		c=c-'A';
	else if (c==' ')
		c = 26;
	else
		assert(0);

	
	count[(int) history[1]][(int) history[0]][(int) c]++;
	// update history
	history[1] = history[0];
	history[0] = c;
}

char tochar(int idx)
{
	if (idx<26)
		return 'A'+idx;
	if (idx==26)
		return '_';
	return '?';
}

// usage: cat *.txt | ./trigrams output.java
int main(int argc, char *argv[])
{
	char c;

    assert(argc==2);

	printf("initializing...\n");
	for (int i=0;i<28;i++)
	{
		for (int j=0;j<28;j++)
		{
			for (int k = 0; k<27; k++)
				count[i][j][k]=0.1; // every combination gets a little epsilon vote
			count[i][j][27]=0;  // except the marginalized entry
		}
	}

	for (int i = 0; i < 2; i++)
		history[i] = 26; // spaces

	printf("counting...\n");

    int nchars = 0;

	while ((c=fgetc(stdin))!=EOF)
	{
        nchars++;

        ///////////////// XXX remove this!!!
        if (1 && !isalpha(c))
            continue;
        /////////////////


		if (isalpha(c))
			update(toupper(c));
		else if (c=='\'')
			continue;
		else
			update(' ');
	}
    printf("counted %d characters\n", nchars);

	printf("marginalizing...\n");
	// marginalizing
	for (int i = 0; i < 27; i++)
	{
		for (int j = 0; j < 27; j++)
		{
			for (int k = 0; k < 27; k++)
			{
				count[27][j][k]  += count[i][j][k];  // ?X*
				count[i][27][k]  += count[i][j][k];  // X?*
				count[27][27][k] += count[i][j][k];  // ??*
			}
		}
	}

	printf("normalizing...\n");

	double globalmin = 0;


	// normalize.
	for (int i = 0; i < 28; i++)
	{
		for (int j = 0; j < 28; j++)
		{
			double sum = 0;
			double min = 999;
			double sum2 = 0;

			// compute some statistics on this probability table
			for (int k = 0; k < 28; k++)
			{
				sum += count[i][j][k];
			}

			// normalize the entries A-Z+space
			for (int k = 0; k < 27; k++)
			{
				count[i][j][k]/=sum;
				count[i][j][k] = log(count[i][j][k]);
				sum2 += count[i][j][k];

				if (count[i][j][k] < min)
					min = count[i][j][k];
				if (count[i][j][k] < globalmin)
					globalmin = count[i][j][k];

			}

			// unknown characters are twice as bad as the worst one.
			// (they're negative...)
			count[i][j][27] = log(1.0/27); //sum2/27; //min*2;
		}
	}


	printf("minimum: %f\n",globalmin);

	printf("writing...\n");

	FILE *outf = fopen(argv[1], "w");

	fprintf(outf, "package decrypto;\n\n");
	fprintf(outf, "public class EnglishTrigrams\n{\n");

	char *tab = "   ";
	
	for (int i = 0; i < 28; i++)
	{
		fprintf(outf, "static final float[][] trigrams%i() { \nreturn new float[][] \n", i);

		fprintf(outf, "%s{\n",tab);
		for (int j = 0; j < 28; j++)
		{
			fprintf(outf, "%s%s// %c%c\n",tab,tab, i+'A', j+'A');

			fprintf(outf, "%s%s{  ",tab,tab);
			for (int k = 0; k < 28; k++)
			{
				float v = count[i][j][k];
				float lv = v;

				fprintf(outf, "%.4ff", lv);
				if (k<27)
					fprintf(outf,", ");
				else
					fprintf(outf," ");
				if ((k&7)==7)
					fprintf(outf,"\n%s%s%s",tab,tab,tab);
			}
			fprintf(outf, "}");
			if (j<27)
				fprintf(outf,",\n");
			else
				fprintf(outf,"\n");
		}
		fprintf(outf, "%s};\n}\n",tab);
/*		if (i<27)
        fprintf(outf,",\n");
		else
        fprintf(outf,"\n");
*/
	}
//	fprintf(outf, "};\n");
	

	fprintf(outf, "static final float trigrams[][][] = new float[][][] {\n");
	for (int i = 0; i < 28; i++)
	{
		fprintf(outf, "trigrams%i()", i);
		if (i<27)
			fprintf(outf,", ");
		else
			fprintf(outf," ");
	}
	fprintf(outf,"};\n");

	fprintf(outf, "}\n");

	fclose(outf);

    ////////////////////////////////////////
    // statistics

    if (0) {
        double max = 99999999999;
        
        // find the 100 most probable trigrams (given the first two
        // letters, the most probable 3rd character)
        for (int iter=0;iter<100;iter++)
        {
            int bi=0, bj=0, bk=0;

            double min = -9999;

            for (int i = 0; i < 28; i++)
            {
                for (int j = 0; j < 28; j++)
                {
                    for (int k = 0; k < 28; k++)
                    {
                        if (count[i][j][k]<max && count[i][j][k]>min)
                        {
                            min = count[i][j][k];
                            bi=i; bj=j; bk=k;
                        }
                    }
                }
            }

            printf("%c%c%c : %f\n",tochar(bi), tochar(bj), tochar(bk), min);
            max = min;
        }

        for (int i = 0; i < 27; i++)
        {
            printf("%c : %10f %10f %10f\n", tochar(i), count[27][27][i], count[27][i][27], count[i][27][27]);
        }
    }
}


